@@ -102,6 +102,16 @@ add_builtin("length", input_types={"x": vec3}, value_type=float, group="Vector M
     doc="Compute the length of a 3d vector.")
 add_builtin("length", input_types={"x": vec4}, value_type=float, group="Vector Math",
     doc="Compute the length of a 4d vector.")
+add_builtin("length", input_types={"x": quat}, value_type=float, group="Vector Math",
+    doc="Compute the length of a quaternion.")
+add_builtin("length_sq", input_types={"x": vec2}, value_type=float, group="Vector Math",
+    doc="Compute the squared length of a 2d vector.")
+add_builtin("length_sq", input_types={"x": vec3}, value_type=float, group="Vector Math",
+    doc="Compute the squared length of a 3d vector.")
+add_builtin("length_sq", input_types={"x": vec4}, value_type=float, group="Vector Math",
+    doc="Compute the squared length of a 4d vector.")
+add_builtin("length_sq", input_types={"x": quat}, value_type=float, group="Vector Math",
+    doc="Compute the squared length of a quaternion.")
 add_builtin("normalize", input_types={"x": vec2}, value_type=vec2, group="Vector Math",
     doc="Compute the normalized value of x, if length(x) is 0 then the zero vector is returned.")
 add_builtin("normalize", input_types={"x": vec3}, value_type=vec3, group="Vector Math",
@@ -134,6 +144,13 @@ add_builtin("determinant", input_types={"m": mat33}, value_type=float, group="Ve
 add_builtin("determinant", input_types={"m": mat44}, value_type=float, group="Vector Math",
     doc="Return the determinant of the matrix m")
 
+add_builtin("trace", input_types={"m": mat22}, value_type=float, group="Vector Math",
+    doc="Return the trace of the matrix m")
+add_builtin("trace", input_types={"m": mat33}, value_type=float, group="Vector Math",
+    doc="Return the trace of the matrix m")
+add_builtin("trace", input_types={"m": mat44}, value_type=float, group="Vector Math",
+    doc="Return the trace of the matrix m")
+
 add_builtin("diag", input_types={"d": vec2}, value_type=mat22, group="Vector Math",
     doc="Returns a matrix with the components of the vector d on the diagonal")
 add_builtin("diag", input_types={"d": vec3}, value_type=mat33, group="Vector Math",
@@ -154,15 +171,12 @@ add_builtin("cw_div", input_types={"x": vec3, "y": vec3}, value_type=vec3, group
 add_builtin("cw_div", input_types={"x": vec4, "y": vec4}, value_type=vec4, group="Vector Math",
     doc="Component wise division of two 4d vectors.")
 
-# type construtors for compute types (int, float)
-for t in scalar_types:
-    add_builtin("int", input_types={"x": t}, value_type=int, doc="Construct an 32-bit signed integer variable, larger precision types will be truncated.", hidden=True, group="Scalar Math", export=False)
-    add_builtin("float", input_types={"x": t}, value_type=float, doc="Construct a 32-bit floating point variable, larger precision types will be truncated.", hidden=True, group="Scalar Math", export=False)
+# type construtors between all storage / compute types 
+scalar_types_all = [*scalar_types, int, float]
+for t in scalar_types_all:
+    for u in scalar_types_all:
+        add_builtin(t.__name__, input_types={"u": u}, value_type=t, doc="", hidden=True, group="Scalar Math", export=False)
 
-# type construtors for storage types 
-for t in scalar_types:
-    add_builtin(t.__name__, input_types={"x": int}, value_type=t, doc="", hidden=True, group="Scalar Math", export=False)
-    add_builtin(t.__name__, input_types={"x": float}, value_type=t, doc="", hidden=True, group="Scalar Math", export=False)
 
 add_builtin("vec2", input_types={}, value_type=vec2, doc="Construct a zero-initialized 2d vector.", group="Vector Math", export=False)
 add_builtin("vec2", input_types={"x": float, "y": float }, value_type=vec2, doc="Construct a 2d vector with compontents x, y.", group="Vector Math", export=False)
@@ -211,6 +225,8 @@ add_builtin("quat_identity", input_types={}, value_type=quat, group="Quaternion
     doc="Construct an identity quaternion with zero imaginary part and real part of 1.0")
 add_builtin("quat_from_axis_angle", input_types={"axis": vec3, "angle": float}, value_type=quat, group="Quaternion Math",
     doc="Construct a quaternion representing a rotation of angle radians around the given axis.")
+add_builtin("quat_to_axis_angle", input_types={"q": quat, "axis": vec3, "angle": float}, value_type=None, group="Quaternion Math",
+    doc="Extract the rotation axis and angle radians a quaternion represents.")
 add_builtin("quat_from_matrix", input_types={"m": mat33}, value_type=quat, group="Quaternion Math",
     doc="Construct a quaternion from a 3x3 matrix.")
 add_builtin("quat_rpy", input_types={"roll": float, "pitch": float, "yaw": float}, value_type=quat, group="Quaternion Math",
@@ -221,6 +237,10 @@ add_builtin("quat_rotate", input_types={"q": quat, "p": vec3}, value_type=vec3,
     doc="Rotate a vector by a quaternion.")
 add_builtin("quat_rotate_inv", input_types={"q": quat, "p": vec3}, value_type=vec3, group="Quaternion Math",
     doc="Rotate a vector the inverse of a quaternion.")
+add_builtin("rotate_rodriguez", input_types={"r": vec3, "x": vec3}, value_type=vec3, group="Quaternion Math",
+    doc="Rotate the vector x by the rotator r encoding rotation axis and angle radians.")
+add_builtin("quat_slerp", input_types={"q0": quat, "q1": quat, "t": float}, value_type=quat, group="Quaternion Math",
+    doc="Linearly interpolate between two quaternions.")
 add_builtin("quat_to_matrix", input_types={"q": quat}, value_type=mat33, group="Quaternion Math",
     doc="Convert a quaternion to a 3x3 rotation matrix.")
 
@@ -370,6 +390,26 @@ add_builtin("mlp", input_types={"weights": array(dtype=float, ndim=2), "bias": a
 #---------------------------------
 # Geometry
 
+add_builtin("bvh_query_aabb", input_types={"id": uint64, "lower": vec3, "upper": vec3}, value_type=bvh_query_t, group="Geometry",
+    doc="""Construct an axis-aligned bounding box query against a bvh object. This query can be used to iterate over all bounds
+   inside a bvh. Returns an object that is used to track state during bvh traversal.
+    
+   :param id: The bvh identifier
+   :param lower: The lower bound of the bounding box in bvh space
+   :param upper: The upper bound of the bounding box in bvh space""")
+
+add_builtin("bvh_query_ray", input_types={"id": uint64, "start": vec3, "dir": vec3}, value_type=bvh_query_t, group="Geometry",
+    doc="""Construct a ray query against a bvh object. This query can be used to iterate over all bounds
+   that intersect the ray. Returns an object that is used to track state during bvh traversal.
+    
+   :param id: The bvh identifier
+   :param start: The start of the ray in bvh space
+   :param dir: The direction of the ray in bvh space""")
+
+add_builtin("bvh_query_next", input_types={"query": bvh_query_t, "index": int}, value_type=bool, group="Geometry",
+    doc="""Move to the next bound returned by the query. The index of the current bound is stored in ``index``, returns ``False``
+   if there are no more overlapping bound.""")
+
 add_builtin("mesh_query_point", input_types={"id": uint64, "point": vec3, "max_dist": float, "inside": float, "face": int, "bary_u": float, "bary_v": float}, value_type=bool, group="Geometry",
     doc="""Computes the closest point on the mesh with identifier `id` to the given point in space. Returns ``True`` if a point < ``max_dist`` is found.
 
@@ -428,6 +468,10 @@ add_builtin("hash_grid_point_id", input_types={"id": uint64, "index": int}, valu
 add_builtin("intersect_tri_tri", input_types={"v0": vec3, "v1": vec3, "v2": vec3, "u0": vec3, "u1": vec3, "u2": vec3}, value_type=int, group="Geometry", 
     doc="Tests for intersection between two triangles (v0, v1, v2) and (u0, u1, u2) using Moller's method. Returns > 0 if triangles intersect.")
 
+
+add_builtin("mesh_get", input_types={"id": uint64}, value_type=Mesh, group="Geometry",
+    doc="""Retrieves the mesh given its index.""")
+
 add_builtin("mesh_eval_face_normal", input_types={"id": uint64, "face": int}, value_type=vec3, group="Geometry",
     doc="""Evaluates the face normal the mesh given a face index.""")
 
@@ -474,18 +518,27 @@ add_builtin("volume_sample_f", input_types={"id": uint64, "uvw": vec3, "sampling
 add_builtin("volume_lookup_f", input_types={"id": uint64, "i": int, "j": int, "k": int}, value_type=float, group="Volumes",
     doc="""Returns the value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value""")
 
+add_builtin("volume_store_f", input_types={"id": uint64, "i": int, "j": int, "k": int, "value": float}, group="Volumes",
+    doc="""Store the value at voxel with coordinates ``i``, ``j``, ``k``.""")
+
 add_builtin("volume_sample_v", input_types={"id": uint64, "uvw": vec3, "sampling_mode": int}, value_type=vec3, group="Volumes",
     doc="""Sample the vector volume given by ``id`` at the volume local-space point ``uvw``. Interpolation should be ``wp.Volume.CLOSEST``, or ``wp.Volume.LINEAR.``""")
 
 add_builtin("volume_lookup_v", input_types={"id": uint64, "i": int, "j": int, "k": int}, value_type=vec3, group="Volumes",
     doc="""Returns the vector value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value""")
 
+add_builtin("volume_store_v", input_types={"id": uint64, "i": int, "j": int, "k": int, "value": vec3}, group="Volumes",
+    doc="""Store the value at voxel with coordinates ``i``, ``j``, ``k``.""")
+
 add_builtin("volume_sample_i", input_types={"id": uint64, "uvw": vec3}, value_type=int, group="Volumes",
     doc="""Sample the int32 volume given by ``id`` at the volume local-space point ``uvw``. """)
 
 add_builtin("volume_lookup_i", input_types={"id": uint64, "i": int, "j": int, "k": int}, value_type=int, group="Volumes",
     doc="""Returns the int32 value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value""")
 
+add_builtin("volume_store_i", input_types={"id": uint64, "i": int, "j": int, "k": int, "value": int}, group="Volumes",
+    doc="""Store the value at voxel with coordinates ``i``, ``j``, ``k``.""")
+
 add_builtin("volume_index_to_world", input_types={"id": uint64, "uvw": vec3}, value_type=vec3, group="Volumes",
     doc="""Transform a point defined in volume index space to world space given the volume's intrinsic affine transformation.""")
 add_builtin("volume_world_to_index", input_types={"id": uint64, "xyz": vec3}, value_type=vec3, group="Volumes",
@@ -558,11 +611,11 @@ add_builtin("pnoise", input_types={"state": uint32, "xyzt": vec4, "px": int, "py
     doc="Periodic Perlin-style noise in 4d.")
 
 add_builtin("curlnoise", input_types={"state": uint32, "xy": vec2}, value_type=vec2, group="Random",
-    doc="Divergence-free vector field based on the gradient of a Perlin noise function.")
+    doc="Divergence-free vector field based on the gradient of a Perlin noise function.", missing_grad=True)
 add_builtin("curlnoise", input_types={"state": uint32, "xyz": vec3}, value_type=vec3, group="Random",
-    doc="Divergence-free vector field based on the curl of three Perlin noise functions.")
+    doc="Divergence-free vector field based on the curl of three Perlin noise functions.", missing_grad=True)
 add_builtin("curlnoise", input_types={"state": uint32, "xyzt": vec4}, value_type=vec3, group="Random",
-    doc="Divergence-free vector field based on the curl of three Perlin noise functions.")
+    doc="Divergence-free vector field based on the curl of three Perlin noise functions.", missing_grad=True)
 
 # note printf calls directly to global CRT printf (no wp:: namespace prefix)
 add_builtin("printf", input_types={}, namespace="", variadic=True, group="Utility",
@@ -703,28 +756,48 @@ add_builtin("atomic_sub", input_types={"a": array(dtype=Any), "i": int, "j": int
 add_builtin("atomic_sub", input_types={"a": array(dtype=Any), "i": int, "j": int, "k":int, "value": Any}, value_func=atomic_op_value_type, doc="Atomically subtract ``value`` onto the array at location given by indices.", group="Utility", skip_replay=True)
 add_builtin("atomic_sub", input_types={"a": array(dtype=Any), "i": int, "j": int, "k":int, "l": int, "value": Any}, value_func=atomic_op_value_type, doc="Atomically subtract ``value`` onto the array at location given by indices.", group="Utility", skip_replay=True)
 
+add_builtin("atomic_min", input_types={"a": array(dtype=Any), "i": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_min", input_types={"a": array(dtype=Any), "i": int, "j": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_min", input_types={"a": array(dtype=Any), "i": int, "j": int, "k": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_min", input_types={"a": array(dtype=Any), "i": int, "j": int, "k": int, "l": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+
+add_builtin("atomic_max", input_types={"a": array(dtype=Any), "i": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_max", input_types={"a": array(dtype=Any), "i": int, "j": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_max", input_types={"a": array(dtype=Any), "i": int, "j": int, "k":int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+add_builtin("atomic_max", input_types={"a": array(dtype=Any), "i": int, "j": int, "k":int, "l": int, "value": Any}, value_func=atomic_op_value_type, doc="Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.", group="Utility", skip_replay=True)
+
+
 # used to index into builtin types, i.e.: y = vec3[1]
-add_builtin("index", input_types={"a": vec2, "i": int}, value_type=float,  group="Utility")
-add_builtin("index", input_types={"a": vec3, "i": int}, value_type=float,  group="Utility")
-add_builtin("index", input_types={"a": vec4, "i": int}, value_type=float,  group="Utility")
-add_builtin("index", input_types={"a": quat, "i": int}, value_type=float,  group="Utility")
+add_builtin("index", input_types={"a": vec2, "i": int}, value_type=float, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": vec3, "i": int}, value_type=float, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": vec4, "i": int}, value_type=float, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": quat, "i": int}, value_type=float, hidden=True, group="Utility")
 
-add_builtin("index", input_types={"a": mat22, "i": int}, value_type=vec2,  group="Utility")
-add_builtin("index", input_types={"a": mat22, "i": int, "j": int}, value_type=float,  group="Utility")
+add_builtin("index", input_types={"a": mat22, "i": int}, value_type=vec2, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": mat22, "i": int, "j": int}, value_type=float, hidden=True, group="Utility")
 
-add_builtin("index", input_types={"a": mat33, "i": int}, value_type=vec3,  group="Utility")
-add_builtin("index", input_types={"a": mat33, "i": int, "j": int}, value_type=float,  group="Utility")
+add_builtin("index", input_types={"a": mat33, "i": int}, value_type=vec3, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": mat33, "i": int, "j": int}, value_type=float, hidden=True, group="Utility")
 
-add_builtin("index", input_types={"a": mat44, "i": int}, value_type=vec4,  group="Utility")
-add_builtin("index", input_types={"a": mat44, "i": int, "j": int}, value_type=float,  group="Utility")
+add_builtin("index", input_types={"a": mat44, "i": int}, value_type=vec4, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": mat44, "i": int, "j": int}, value_type=float, hidden=True, group="Utility")
 
+add_builtin("index", input_types={"a": spatial_matrix, "i": int, "j": int}, value_type=float, hidden=True, group="Utility")
+
+add_builtin("index", input_types={"a": spatial_vector, "i": int}, value_type=float, hidden=True, group="Utility")
+add_builtin("index", input_types={"a": transform, "i": int}, value_type=float, hidden=True, group="Utility")
+
+add_builtin("index", input_types={"s": shape_t, "i": int}, value_type=int, hidden=True, group="Utility")
 
 for t in scalar_types + vector_types:
     add_builtin("expect_eq", input_types={"arg1": t, "arg2": t}, value_type=None, doc="Prints an error to stdout if arg1 and arg2 are not equal", group="Utility")
 
-    if type_is_int(t) == False:
+for t in compute_types + vector_types:
+    if not type_is_int(t):
         add_builtin("lerp", input_types={"a": t, "b": t, "t": float}, value_type=t, doc="Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``", group="Utility")
 
+add_builtin("smoothstep", input_types={"edge0": float, "edge1": float, "x": float}, value_type=float, doc="Smoothly interpolate between two values edge0 and edge1 using a factor x, and return a result between 0 and 1 using a cubic Hermite interpolation after clamping", group="Utility")
+
 # fuzzy compare for float values
 add_builtin("expect_near", input_types={"arg1": float, "arg2": float, "tolerance": float}, value_type=None, doc="Prints an error to stdout if arg1 and arg2 are not closer than tolerance in magnitude", group="Utility")
 add_builtin("expect_near", input_types={"arg1": vec3, "arg2": vec3, "tolerance": float}, value_type=None, doc="Prints an error to stdout if any element of arg1 and arg2 are not closer than tolerance in magnitude", group="Utility")
