@@ -42,6 +42,11 @@ WP_API void builtin_skew_vec3(vec3 x, mat33* ret) { *ret = wp::skew(x); }
 WP_API void builtin_length_vec2(vec2 x, float* ret) { *ret = wp::length(x); }
 WP_API void builtin_length_vec3(vec3 x, float* ret) { *ret = wp::length(x); }
 WP_API void builtin_length_vec4(vec4 x, float* ret) { *ret = wp::length(x); }
+WP_API void builtin_length_quat(quat x, float* ret) { *ret = wp::length(x); }
+WP_API void builtin_length_sq_vec2(vec2 x, float* ret) { *ret = wp::length_sq(x); }
+WP_API void builtin_length_sq_vec3(vec3 x, float* ret) { *ret = wp::length_sq(x); }
+WP_API void builtin_length_sq_vec4(vec4 x, float* ret) { *ret = wp::length_sq(x); }
+WP_API void builtin_length_sq_quat(quat x, float* ret) { *ret = wp::length_sq(x); }
 WP_API void builtin_normalize_vec2(vec2 x, vec2* ret) { *ret = wp::normalize(x); }
 WP_API void builtin_normalize_vec3(vec3 x, vec3* ret) { *ret = wp::normalize(x); }
 WP_API void builtin_normalize_vec4(vec4 x, vec4* ret) { *ret = wp::normalize(x); }
@@ -56,6 +61,9 @@ WP_API void builtin_inverse_mat44(mat44 m, mat44* ret) { *ret = wp::inverse(m);
 WP_API void builtin_determinant_mat22(mat22 m, float* ret) { *ret = wp::determinant(m); }
 WP_API void builtin_determinant_mat33(mat33 m, float* ret) { *ret = wp::determinant(m); }
 WP_API void builtin_determinant_mat44(mat44 m, float* ret) { *ret = wp::determinant(m); }
+WP_API void builtin_trace_mat22(mat22 m, float* ret) { *ret = wp::trace(m); }
+WP_API void builtin_trace_mat33(mat33 m, float* ret) { *ret = wp::trace(m); }
+WP_API void builtin_trace_mat44(mat44 m, float* ret) { *ret = wp::trace(m); }
 WP_API void builtin_diag_vec2(vec2 d, mat22* ret) { *ret = wp::diag(d); }
 WP_API void builtin_diag_vec3(vec3 d, mat33* ret) { *ret = wp::diag(d); }
 WP_API void builtin_diag_vec4(vec4 d, mat44* ret) { *ret = wp::diag(d); }
@@ -68,11 +76,14 @@ WP_API void builtin_cw_div_vec4_vec4(vec4 x, vec4 y, vec4* ret) { *ret = wp::cw_
 WP_API void builtin_svd3_mat33_mat33_vec3_mat33(mat33 A, mat33 U, vec3 sigma, mat33 V) { wp::svd3(A, U, sigma, V); }
 WP_API void builtin_quat_identity(quat* ret) { *ret = wp::quat_identity(); }
 WP_API void builtin_quat_from_axis_angle_vec3_float32(vec3 axis, float32 angle, quat* ret) { *ret = wp::quat_from_axis_angle(axis, angle); }
+WP_API void builtin_quat_to_axis_angle_quat_vec3_float32(quat q, vec3 axis, float32 angle) { wp::quat_to_axis_angle(q, axis, angle); }
 WP_API void builtin_quat_from_matrix_mat33(mat33 m, quat* ret) { *ret = wp::quat_from_matrix(m); }
 WP_API void builtin_quat_rpy_float32_float32_float32(float32 roll, float32 pitch, float32 yaw, quat* ret) { *ret = wp::quat_rpy(roll, pitch, yaw); }
 WP_API void builtin_quat_inverse_quat(quat q, quat* ret) { *ret = wp::quat_inverse(q); }
 WP_API void builtin_quat_rotate_quat_vec3(quat q, vec3 p, vec3* ret) { *ret = wp::quat_rotate(q, p); }
 WP_API void builtin_quat_rotate_inv_quat_vec3(quat q, vec3 p, vec3* ret) { *ret = wp::quat_rotate_inv(q, p); }
+WP_API void builtin_rotate_rodriguez_vec3_vec3(vec3 r, vec3 x, vec3* ret) { *ret = wp::rotate_rodriguez(r, x); }
+WP_API void builtin_quat_slerp_quat_quat_float32(quat q0, quat q1, float32 t, quat* ret) { *ret = wp::quat_slerp(q0, q1, t); }
 WP_API void builtin_quat_to_matrix_quat(quat q, mat33* ret) { *ret = wp::quat_to_matrix(q); }
 WP_API void builtin_transform_identity(transform* ret) { *ret = wp::transform_identity(); }
 WP_API void builtin_transform_get_translation_transform(transform t, vec3* ret) { *ret = wp::transform_get_translation(t); }
@@ -88,6 +99,9 @@ WP_API void builtin_spatial_cross_spatial_vector_spatial_vector(spatial_vector a
 WP_API void builtin_spatial_cross_dual_spatial_vector_spatial_vector(spatial_vector a, spatial_vector b, spatial_vector* ret) { *ret = wp::spatial_cross_dual(a, b); }
 WP_API void builtin_spatial_top_spatial_vector(spatial_vector a, vec3* ret) { *ret = wp::spatial_top(a); }
 WP_API void builtin_spatial_bottom_spatial_vector(spatial_vector a, vec3* ret) { *ret = wp::spatial_bottom(a); }
+WP_API void builtin_bvh_query_aabb_uint64_vec3_vec3(uint64 id, vec3 lower, vec3 upper, bvh_query_t* ret) { *ret = wp::bvh_query_aabb(id, lower, upper); }
+WP_API void builtin_bvh_query_ray_uint64_vec3_vec3(uint64 id, vec3 start, vec3 dir, bvh_query_t* ret) { *ret = wp::bvh_query_ray(id, start, dir); }
+WP_API void builtin_bvh_query_next_bvh_query_t_int32(bvh_query_t query, int32 index, bool* ret) { *ret = wp::bvh_query_next(query, index); }
 WP_API void builtin_mesh_query_point_uint64_vec3_float32_float32_int32_float32_float32(uint64 id, vec3 point, float32 max_dist, float32 inside, int32 face, float32 bary_u, float32 bary_v, bool* ret) { *ret = wp::mesh_query_point(id, point, max_dist, inside, face, bary_u, bary_v); }
 WP_API void builtin_mesh_query_ray_uint64_vec3_vec3_float32_float32_float32_float32_float32_vec3_int32(uint64 id, vec3 start, vec3 dir, float32 max_t, float32 t, float32 bary_u, float32 bary_v, float32 sign, vec3 normal, int32 face, bool* ret) { *ret = wp::mesh_query_ray(id, start, dir, max_t, t, bary_u, bary_v, sign, normal, face); }
 WP_API void builtin_mesh_query_aabb_uint64_vec3_vec3(uint64 id, vec3 lower, vec3 upper, mesh_query_aabb_t* ret) { *ret = wp::mesh_query_aabb(id, lower, upper); }
@@ -98,19 +112,24 @@ WP_API void builtin_hash_grid_query_uint64_vec3_float32(uint64 id, vec3 point, f
 WP_API void builtin_hash_grid_query_next_hash_grid_query_t_int32(hash_grid_query_t query, int32 index, bool* ret) { *ret = wp::hash_grid_query_next(query, index); }
 WP_API void builtin_hash_grid_point_id_uint64_int32(uint64 id, int32 index, int* ret) { *ret = wp::hash_grid_point_id(id, index); }
 WP_API void builtin_intersect_tri_tri_vec3_vec3_vec3_vec3_vec3_vec3(vec3 v0, vec3 v1, vec3 v2, vec3 u0, vec3 u1, vec3 u2, int* ret) { *ret = wp::intersect_tri_tri(v0, v1, v2, u0, u1, u2); }
+WP_API void builtin_mesh_get_uint64(uint64 id, Mesh* ret) { *ret = wp::mesh_get(id); }
 WP_API void builtin_mesh_eval_face_normal_uint64_int32(uint64 id, int32 face, vec3* ret) { *ret = wp::mesh_eval_face_normal(id, face); }
 WP_API void builtin_mesh_get_point_uint64_int32(uint64 id, int32 index, vec3* ret) { *ret = wp::mesh_get_point(id, index); }
 WP_API void builtin_mesh_get_velocity_uint64_int32(uint64 id, int32 index, vec3* ret) { *ret = wp::mesh_get_velocity(id, index); }
 WP_API void builtin_mesh_get_index_uint64_int32(uint64 id, int32 index, int* ret) { *ret = wp::mesh_get_index(id, index); }
+WP_API void builtin_closest_point_edge_edge_vec3_vec3_vec3_vec3_float32(vec3 p1, vec3 q1, vec3 p2, vec3 q2, float32 epsilon, vec3* ret) { *ret = wp::closest_point_edge_edge(p1, q1, p2, q2, epsilon); }
 WP_API void builtin_iter_next_range_t(range_t range, int* ret) { *ret = wp::iter_next(range); }
 WP_API void builtin_iter_next_hash_grid_query_t(hash_grid_query_t query, int* ret) { *ret = wp::iter_next(query); }
 WP_API void builtin_iter_next_mesh_query_aabb_t(mesh_query_aabb_t query, int* ret) { *ret = wp::iter_next(query); }
 WP_API void builtin_volume_sample_f_uint64_vec3_int32(uint64 id, vec3 uvw, int32 sampling_mode, float* ret) { *ret = wp::volume_sample_f(id, uvw, sampling_mode); }
 WP_API void builtin_volume_lookup_f_uint64_int32_int32_int32(uint64 id, int32 i, int32 j, int32 k, float* ret) { *ret = wp::volume_lookup_f(id, i, j, k); }
+WP_API void builtin_volume_store_f_uint64_int32_int32_int32_float32(uint64 id, int32 i, int32 j, int32 k, float32 value) { wp::volume_store_f(id, i, j, k, value); }
 WP_API void builtin_volume_sample_v_uint64_vec3_int32(uint64 id, vec3 uvw, int32 sampling_mode, vec3* ret) { *ret = wp::volume_sample_v(id, uvw, sampling_mode); }
 WP_API void builtin_volume_lookup_v_uint64_int32_int32_int32(uint64 id, int32 i, int32 j, int32 k, vec3* ret) { *ret = wp::volume_lookup_v(id, i, j, k); }
+WP_API void builtin_volume_store_v_uint64_int32_int32_int32_vec3(uint64 id, int32 i, int32 j, int32 k, vec3 value) { wp::volume_store_v(id, i, j, k, value); }
 WP_API void builtin_volume_sample_i_uint64_vec3(uint64 id, vec3 uvw, int* ret) { *ret = wp::volume_sample_i(id, uvw); }
 WP_API void builtin_volume_lookup_i_uint64_int32_int32_int32(uint64 id, int32 i, int32 j, int32 k, int* ret) { *ret = wp::volume_lookup_i(id, i, j, k); }
+WP_API void builtin_volume_store_i_uint64_int32_int32_int32_int32(uint64 id, int32 i, int32 j, int32 k, int32 value) { wp::volume_store_i(id, i, j, k, value); }
 WP_API void builtin_volume_index_to_world_uint64_vec3(uint64 id, vec3 uvw, vec3* ret) { *ret = wp::volume_index_to_world(id, uvw); }
 WP_API void builtin_volume_world_to_index_uint64_vec3(uint64 id, vec3 xyz, vec3* ret) { *ret = wp::volume_world_to_index(id, xyz); }
 WP_API void builtin_volume_index_to_world_dir_uint64_vec3(uint64 id, vec3 uvw, vec3* ret) { *ret = wp::volume_index_to_world_dir(id, uvw); }
@@ -122,6 +141,15 @@ WP_API void builtin_randi_uint32_int32_int32(uint32 state, int32 min, int32 max,
 WP_API void builtin_randf_uint32(uint32 state, float* ret) { *ret = wp::randf(state); }
 WP_API void builtin_randf_uint32_float32_float32(uint32 state, float32 min, float32 max, float* ret) { *ret = wp::randf(state, min, max); }
 WP_API void builtin_randn_uint32(uint32 state, float* ret) { *ret = wp::randn(state); }
+WP_API void builtin_sample_triangle_uint32(uint32 state, vec2* ret) { *ret = wp::sample_triangle(state); }
+WP_API void builtin_sample_unit_ring_uint32(uint32 state, vec2* ret) { *ret = wp::sample_unit_ring(state); }
+WP_API void builtin_sample_unit_disk_uint32(uint32 state, vec2* ret) { *ret = wp::sample_unit_disk(state); }
+WP_API void builtin_sample_unit_sphere_surface_uint32(uint32 state, vec3* ret) { *ret = wp::sample_unit_sphere_surface(state); }
+WP_API void builtin_sample_unit_sphere_uint32(uint32 state, vec3* ret) { *ret = wp::sample_unit_sphere(state); }
+WP_API void builtin_sample_unit_hemisphere_surface_uint32(uint32 state, vec3* ret) { *ret = wp::sample_unit_hemisphere_surface(state); }
+WP_API void builtin_sample_unit_hemisphere_uint32(uint32 state, vec3* ret) { *ret = wp::sample_unit_hemisphere(state); }
+WP_API void builtin_sample_unit_square_uint32(uint32 state, vec2* ret) { *ret = wp::sample_unit_square(state); }
+WP_API void builtin_sample_unit_cube_uint32(uint32 state, vec3* ret) { *ret = wp::sample_unit_cube(state); }
 WP_API void builtin_noise_uint32_float32(uint32 state, float32 x, float* ret) { *ret = wp::noise(state, x); }
 WP_API void builtin_noise_uint32_vec2(uint32 state, vec2 xy, float* ret) { *ret = wp::noise(state, xy); }
 WP_API void builtin_noise_uint32_vec3(uint32 state, vec3 xyz, float* ret) { *ret = wp::noise(state, xyz); }
@@ -137,12 +165,17 @@ WP_API void builtin_tid(int* ret) { *ret = wp::tid(); }
 WP_API void builtin_index_vec2_int32(vec2 a, int32 i, float* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_vec3_int32(vec3 a, int32 i, float* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_vec4_int32(vec4 a, int32 i, float* ret) { *ret = wp::index(a, i); }
+WP_API void builtin_index_quat_int32(quat a, int32 i, float* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_mat22_int32(mat22 a, int32 i, vec2* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_mat22_int32_int32(mat22 a, int32 i, int32 j, float* ret) { *ret = wp::index(a, i, j); }
 WP_API void builtin_index_mat33_int32(mat33 a, int32 i, vec3* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_mat33_int32_int32(mat33 a, int32 i, int32 j, float* ret) { *ret = wp::index(a, i, j); }
 WP_API void builtin_index_mat44_int32(mat44 a, int32 i, vec4* ret) { *ret = wp::index(a, i); }
 WP_API void builtin_index_mat44_int32_int32(mat44 a, int32 i, int32 j, float* ret) { *ret = wp::index(a, i, j); }
+WP_API void builtin_index_spatial_matrix_int32_int32(spatial_matrix a, int32 i, int32 j, float* ret) { *ret = wp::index(a, i, j); }
+WP_API void builtin_index_spatial_vector_int32(spatial_vector a, int32 i, float* ret) { *ret = wp::index(a, i); }
+WP_API void builtin_index_transform_int32(transform a, int32 i, float* ret) { *ret = wp::index(a, i); }
+WP_API void builtin_index_shape_t_int32(shape_t s, int32 i, int* ret) { *ret = wp::index(s, i); }
 WP_API void builtin_expect_eq_int8_int8(int8 arg1, int8 arg2) { wp::expect_eq(arg1, arg2); }
 WP_API void builtin_expect_eq_uint8_uint8(uint8 arg1, uint8 arg2) { wp::expect_eq(arg1, arg2); }
 WP_API void builtin_expect_eq_int16_int16(int16 arg1, int16 arg2) { wp::expect_eq(arg1, arg2); }
@@ -164,9 +197,7 @@ WP_API void builtin_expect_eq_quat_quat(quat arg1, quat arg2) { wp::expect_eq(ar
 WP_API void builtin_expect_eq_transform_transform(transform arg1, transform arg2) { wp::expect_eq(arg1, arg2); }
 WP_API void builtin_expect_eq_spatial_vector_spatial_vector(spatial_vector arg1, spatial_vector arg2) { wp::expect_eq(arg1, arg2); }
 WP_API void builtin_expect_eq_spatial_matrix_spatial_matrix(spatial_matrix arg1, spatial_matrix arg2) { wp::expect_eq(arg1, arg2); }
-WP_API void builtin_lerp_float16_float16_float32(float16 a, float16 b, float32 t, float16* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_float32_float32_float32(float32 a, float32 b, float32 t, float32* ret) { *ret = wp::lerp(a, b, t); }
-WP_API void builtin_lerp_float64_float64_float32(float64 a, float64 b, float32 t, float64* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_vec2_vec2_float32(vec2 a, vec2 b, float32 t, vec2* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_vec3_vec3_float32(vec3 a, vec3 b, float32 t, vec3* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_vec4_vec4_float32(vec4 a, vec4 b, float32 t, vec4* ret) { *ret = wp::lerp(a, b, t); }
@@ -177,6 +208,7 @@ WP_API void builtin_lerp_quat_quat_float32(quat a, quat b, float32 t, quat* ret)
 WP_API void builtin_lerp_transform_transform_float32(transform a, transform b, float32 t, transform* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_spatial_vector_spatial_vector_float32(spatial_vector a, spatial_vector b, float32 t, spatial_vector* ret) { *ret = wp::lerp(a, b, t); }
 WP_API void builtin_lerp_spatial_matrix_spatial_matrix_float32(spatial_matrix a, spatial_matrix b, float32 t, spatial_matrix* ret) { *ret = wp::lerp(a, b, t); }
+WP_API void builtin_smoothstep_float32_float32_float32(float32 edge0, float32 edge1, float32 x, float* ret) { *ret = wp::smoothstep(edge0, edge1, x); }
 WP_API void builtin_expect_near_float32_float32_float32(float32 arg1, float32 arg2, float32 tolerance) { wp::expect_near(arg1, arg2, tolerance); }
 WP_API void builtin_expect_near_vec3_vec3_float32(vec3 arg1, vec3 arg2, float32 tolerance) { wp::expect_near(arg1, arg2, tolerance); }
 WP_API void builtin_add_int32_int32(int32 x, int32 y, int* ret) { *ret = wp::add(x, y); }
