@@ -139,5 +139,10 @@ class Tape:
 
         for a, g in self.gradients.items():
             if a not in self.const_gradients:
-                g.zero_()
+                if isinstance(a, wp.codegen.StructInstance):
+                    for name in g._struct_.vars:
+                        if isinstance(g._struct_.vars[name].type, wp.array):
+                            getattr(g, name).zero_()
+                else:
+                    g.zero_()
 
